#ifndef __PATHALGEBRASnBASICTYPES__ 
#define __PATHALGEBRASnBASICTYPES__ 

#include <iostream>
#include <iomanip>
#include <sstream>
#include <string>
#include <set>
#include <vector>
#include <limits>
#include <regex>

using namespace std;

const double DoublePrecision = 1E-12; //10^-12

int PreciseDigits(const double precision = DoublePrecision) {
	int n = 0; double ten = 1;
	do { ten *= 10; ++n;
	} while (precision * ten < 0.99);
	return n;
}

const double DoublePrecisionDigits = PreciseDigits(DoublePrecision);
const double DoublePrecisionDigitsForOutput = 0;

inline bool DoubleEqual(const double &d1, const double &d2, const double precision = DoublePrecision) {
	if (d1 == d2) return true; 
	if (abs(d1 - d2) < precision) return true; 
	return false;
}


class P1Type {
//Accessible sets
	bool element;
public:
	const static P1Type zero;
	const static P1Type one;
	P1Type(bool b = zero.element) : element (b) {} ;
	const P1Type& operator+ (const P1Type &s) const {
		if (element >= s.element) return (*this);
		return s;
	}
	const P1Type& operator* (const P1Type &s) const {
		if (element <= s.element) return (*this);
		return s;
	}
	bool operator==(const P1Type &s) const {
		if (element == s.element) return true;
		return false;
	}
	bool operator<=(const P1Type &s) const {
		return   ( (element + s.element) == s.element ); 
	}
	string ToString() const {
		if (element == true) return string("1");
		return string("0");
	}
	string ToDelimitedString(char delim = ' ') const {
		return ToString();
	}
	void FromString(string &s) {
		auto it =  find_if( s.begin() , s.end() , [](char ch){ return !isspace<char>(ch , locale::classic() ) ; } );
  		s.erase( s.begin() , it);
		istringstream is (s);
		ostringstream os;
		is >> element;
		if (is.fail()) exit(1);
		if (is.tellg() > 0) s = s.substr(is.tellg());
		else s = "";
	}
	void FromDelimitedString(string &s, char delim = ' ') {
		FromString(s);
	}
	const P1Type& Zero() const {
		return zero;
	}
	const P1Type& One() const {
		return one;
	}
};

inline ostream& operator<< (ostream &os, const P1Type &s) {
	os << s.ToString();
	return os;
}

template <class Type>
class P2Type {
//Shortest path
	Type element;
public:
	static const P2Type zero;
	static const P2Type one;
	P2Type(Type d = zero.element) : element (d) {} ;
	const P2Type& operator+ (const P2Type &s) const {
		if (element <= s.element) return (*this);
		return s;
	}
	const P2Type& operator* (const P2Type &s) const {
		static P2Type *r = new P2Type(); 
		r->element = element + s.element;
		return *r; 
	}
	bool operator==(const P2Type &s) const {
		if (element == s.element) return true;
		return false;
	}
	bool operator<=(const P2Type &s) const {
		return   (element >= s.element); 
	}
	inline bool lex_LE(const P2Type &s) const {
		return   (element < s.element); 
	}
	string ToString() const {
		ostringstream os;
		os << element;
		return os.str();
	}
	string ToDelimitedString(char delim = ' ') const {
		return ToString();
	}

	void FromString(string &s) {
		auto it =  find_if( s.begin() , s.end() , [](char ch){ return !isspace<char>(ch , locale::classic() ) ; } );
  		s.erase( s.begin() , it);
		istringstream is (s);
		ostringstream os;
		is >> element;
		if (is.fail()) exit(1);
		if (is.tellg() > 0) s = s.substr(is.tellg());
		else s = "";
	}

	void FromDelimitedString(string &s, char delim = ' ') {
		FromString(s);
	}

	const Type& GetElement() const {
		return element;
	}
	
	void SetElement(const Type& d)  {
		element = d;
	}

	const P2Type& Zero() const {
		return zero;
	}
	const P2Type& One() const {
		return one;
	}
};

template<>
inline bool P2Type<double>::operator==(const P2Type &s) const {
	return DoubleEqual(element,s.element);
}

template<>
string P2Type<double>::ToString() const {
	ostringstream os;
	os << fixed << setprecision(DoublePrecisionDigitsForOutput);
	os << element;
	return os.str();
}


template<class Type>
using PSType = P2Type<Type>;

template<class Type>
inline ostream& operator<< (ostream &os, const P2Type<Type> &s) {
	os << s.ToString();
	return os;
}

template<class Type>
class P3Type {
//Longest (critical) path
	Type element;
public:
	const static P3Type zero;
	const static P3Type one;
	P3Type(Type d = 0.0) {
		element = d;
	}
	const P3Type& operator+ (const P3Type &s) const {
		if (element >= s.element) return (*this);
		return s;
	}
	const P3Type& operator* (const P3Type &s) const {
		static P3Type *r = new P3Type(); 
		r->element = element + s.element;
		return *r; 
	}
	bool operator==(const P3Type &s) const {
		if (element == s.element) return true;
		return false;
	}
	bool operator<=(const P3Type &s) const {
		return   (element <= s.element); 
	}
	bool lex_LE(const P3Type &s) const {
		return   (element <= s.element); 
	}
	string ToString() const {
		ostringstream os;
		os << element;
		return os.str();
	}
	string ToDelimitedString(char delim = ' ') const {
		return ToString();
	}
	void FromString(string &s) {
		auto it =  find_if( s.begin() , s.end() , [](char ch){ return !isspace<char>(ch , locale::classic() ) ; } );
  		s.erase( s.begin() , it);
		istringstream is (s);
		ostringstream os;
		is >> element;
		if (is.fail()) exit(1);
		if (is.tellg() > 0) s = s.substr(is.tellg());
		else s = "";
	}
	void FromDelimitedString(string &s, char delim = ' ') {
		FromString(s);
	}
	const P3Type& Zero() const {
		return zero;
	}
	const P3Type& One() const {
		return one;
	}
};


template<>
inline bool P3Type<double>::operator==(const P3Type &s) const {
	return DoubleEqual(element,s.element);
}

template<>
string P3Type<double>::ToString() const {
	ostringstream os;
	os << fixed << setprecision(DoublePrecisionDigitsForOutput);
	os << element;
	return os.str();
}

template<class Type>
using PLType = P3Type<Type>;

template<class Type>
inline ostream& operator<< (ostream &os, const P3Type<Type> &s) {
	os << s.ToString();
	return os;
}

class P4Type {
//Most reliable path
	double element; // in [0,1]
public:
	const static P4Type zero;
	const static P4Type one;
	P4Type(double d = 0.0) : element (d) {
		if (d < 0.0 || d > 1.0) element = 0.0;
	}
	const P4Type& operator+ (const P4Type &s) const {
		if (element >= s.element) return (*this);
		return s;
	}
	const P4Type& operator* (const P4Type &s) const {
		static P4Type *r = new P4Type(); 
		r->element = element * s.element;
		return *r; 
	}
	bool operator==(const P4Type &s) const {
		return DoubleEqual(element,s.element);
	}
	inline bool operator<=(const P4Type &s) const {
		return   (element <= s.element); 
	}
	inline bool lex_LE(const P4Type &s) const {
		return   (element < s.element); 
	}
	string ToString() const {
		ostringstream os;
		os << fixed << setprecision(DoublePrecisionDigitsForOutput);
		os << element;
		return os.str();
	}
	string ToDelimitedString(char delim = ' ') const {
		return ToString();
	}

	void FromString(string &s) {
		auto it =  find_if( s.begin() , s.end() , [](char ch){ return !isspace<char>(ch , locale::classic() ) ; } );
  		s.erase( s.begin() , it);
		istringstream is (s);
		is >> element;
		if (is.fail()) exit(1);
		if (is.tellg() > 0) s = s.substr(is.tellg());
		else s = "";
	}

	void FromDelimitedString(string &s, char delim = ' ') {
		FromString(s);
	}

	const P4Type& Zero() const {
		return zero;
	}
	const P4Type& One() const {
		return one;
	}
};


typedef P4Type PRType;

inline ostream& operator<< (ostream &os, const P4Type &s) {
	os << s.ToString();
	return os;
}

template<class Type>
class P5Type {
//Paths of greatest capacity
	Type element; // in [0,+inf>
public:
	const static P5Type zero;
	const static P5Type one;
	P5Type(Type d = 0.0) : element (d) {
		if (d < 0.0) element = 0.0;
	}
	const P5Type& operator+ (const P5Type &s) const {
		if (element >= s.element) return (*this);
		return s;
	}
	const P5Type& operator* (const P5Type &s) const {
		if (element <= s.element) return (*this);
		return s;
	}
	bool operator==(const P5Type &s) const {
		if (element == s.element) return true;
		return false;
	}
	bool operator<=(const P5Type &s) const {
		return   (element <= s.element); 
	}
	string ToString() const {
		ostringstream os;
		os << fixed << setprecision(DoublePrecisionDigitsForOutput);
		os << element;
		return os.str();
	}
	string ToDelimitedString(char delim = ' ') const {
		return ToString();
	}
	void FromString(string &s) {
		auto it =  find_if( s.begin() , s.end() , [](char ch){ return !isspace<char>(ch , locale::classic() ) ; } );
  		s.erase( s.begin() , it);
		istringstream is (s);
		ostringstream os;
		is >> element;
		if (is.fail()) exit(1);
		if (is.tellg() > 0) s = s.substr(is.tellg());
		else s = "";
	}
	void FromDelimitedString(string &s, char delim = ' ') {
		FromString(s);
	}
	const P5Type& Zero() const {
		return zero;
	}
	const P5Type& One() const {
		return one;
	}
};


template<>
inline bool P5Type<double>::operator==(const P5Type &s) const {
	return DoubleEqual(element,s.element);
}

template<>
string P5Type<double>::ToString() const {
	ostringstream os;
	os << fixed << setprecision(DoublePrecisionDigitsForOutput);
	os << element;
	return os.str();
}


template<class Type>
using PCType = P5Type<Type>;

template<class Type>
inline ostream& operator<< (ostream &os, const P5Type<Type> &s) {
	os << s.ToString();
	return os;
}

template<class Type>
class LetterType {
public:
	Type letter;
	LetterType() {}
	LetterType(const Type &c) : letter(c) {}
	string ToString() const {
		ostringstream os;
		os << letter;
		return os.str();
	}
	string ToDelimitedString() const {
		return ToString();
	}
	void FromString(string &s)  {
		auto it =  find_if( s.begin() , s.end() , [](char ch){ return !isspace<char>(ch , locale::classic() ) ; } );
  		s.erase( s.begin() , it);
  		istringstream is (s);
  		is >> letter;
  		ostringstream os;
		os << letter;
		int n = os.str().size() ;
		s = s.substr(n);
	}
	bool operator==(const LetterType& s) const { return (letter == s.letter); }
	bool operator!=(const LetterType& s) const { return (letter != s.letter); }
	bool operator< (const LetterType& s) const { return (letter <  s.letter);	}
	bool operator<=(const LetterType& s) const { return (letter <= s.letter);	}
	bool operator> (const LetterType& s) const { return (letter >  s.letter);	}
	bool operator>=(const LetterType& s) const { return (letter >= s.letter);	}
};

template<class Type>
inline ostream& operator<< (ostream &os, const LetterType<Type> &s) {
	os << s.ToString();
	return os;
}

template<class Type>
class Alphabet {
public:
	vector<Type> alphabet;
	void Add(const Type &letter) {
		for (int i=0; i<alphabet.size(); ++i) {
			if (alphabet[i] == letter) return;
		}
		alphabet.push_back(letter);
		Sort();
	}
	bool Has(const Type &letter) {
		for (int i=0; i<alphabet.size(); ++i) {
			if (alphabet[i] == letter) return true;
		}
		return false;
	}
	void Sort() {
		for (int i=0; i<alphabet.size()-1; ++i) {
			for (int j=i+1; j<alphabet.size(); ++j) {
				if (alphabet[i] > alphabet[j]) {
					Type tmp = alphabet[i];
					alphabet[i] = alphabet[j];
					alphabet[j] = tmp;
				}
			}
		}
	}
	Type & operator[] (int i) {
		return alphabet[i];
	}
	string ToString() const {
		string s;
		s = "{";
		for (int i=0; i<alphabet.size(); ++i) {
			s += " " + alphabet[i].ToString() + " ,";
		}
		if (s[s.size()-1] == ',') s[s.size()-1] = '}';
		else s += " }";
		return s;
	}
	void FromString(string &s)  {
		auto it =  find_if( s.begin() , s.end() , [](char ch){ return !isspace<char>(ch , locale::classic() ) ; } );
  		s.erase( s.begin() , it);
  		istringstream is (s);
  		Type letter;
  		is >> letter;
  		ostringstream os;
		os << letter;
		int n = os.str().size() ;
		s = s.substr(n);
	}

	void FromDelimitedString(string &s, char delim = ' ') {
		auto it =  find_if( s.begin() , s.end() , [](char ch){ return !isspace<char>(ch , locale::classic() ) ; } );
  		s.erase( s.begin() , it);
  		istringstream is (s);
  		Type letter;
  		is >> letter;
  		ostringstream os;
		os << letter;
		int n = os.str().size() ;
		s = s.substr(n);
	}
	bool operator==(const Alphabet& s) const { return (alphabet == s.alphabet); }
	bool operator!=(const Alphabet& s) const { return (alphabet != s.alphabet); }
};

template<class Type>
inline ostream& operator<< (ostream &os, const Alphabet<Type> &s) {
	os << s.ToString();
	return os;
}

template<class Type>
class WordType {
public:
	Alphabet<Type> alphabet; 
	vector<Type> word;
	WordType() {}
	WordType(const Type &letter) {word.push_back(letter);}
	WordType(const Alphabet<Type> &a) : alphabet (a) {}

	string ToString() const {
		string s;
		s = "<";
		for (int i=0; i<word.size(); ++i) {
			s += word[i].ToString() + "_";
		}
		if (s[s.size()-1] == '_') s[s.size()-1] = '>';
		else s += ">";
		return s;
	}

	string ToDelimitedString(char delim = ' ') const {
		string s;
		for (int i=0; i<word.size(); ++i) {
			s += word[i].ToDelimitedString() + delim;
		}
		return s;
	}

	void FromString(string &s)  {
		int i;
		auto it =  find_if( s.begin() , s.end() , [](char ch){ return !isspace<char>(ch , locale::classic() ) ; } );
  		s.erase( s.begin() , it);
  		if (s[0] == '<') s = s.substr(1);
  		else exit(0);
		Type letter;
		string letterstring;
		
		while (true) {
	  		for (i=0; i<s.size(); ++i) {
	  			if (s[i] == '_' || s[i] == '>' || s[i] == ' ') break;
			}
			letterstring = s.substr(0,i);
			letter.FromString(letterstring);
			word.push_back(letter) ;
			s = s.substr(i);
			it =  find_if( s.begin() , s.end() , [](char ch){ return !isspace<char>(ch , locale::classic() ) ; } );
	  		s.erase( s.begin() , it);
	  		if (s[0] == '>') {
				s = s.substr(1);
				break;
			}
			else if (s[0] == '_') {
				s = s.substr(1);
				continue;
			}
	  		else exit(0);
		}
	}
	const WordType& operator*(const WordType& s) const { 
		static WordType *r = new WordType(alphabet);
		r->word = word;
		r->word.insert(r->word.end(), s.word.begin(), s.word.end());
		return (*r); 
	}
	
	bool operator==(const WordType& s) const { return (word == s.word); }
	bool operator!=(const WordType& s) const { return (word != s.word); }
	bool operator< (const WordType& s) const { return (word <  s.word);	}
	bool operator<=(const WordType& s) const { return (word <= s.word);	}
	bool operator> (const WordType& s) const { return (word >  s.word);	}
	bool operator>=(const WordType& s) const { return (word >= s.word);	}
	LetterType<Type> & operator[] (int i) {	
		return word[i];	
	}
	void Add(const Type &letter) {
		if (alphabet.Has(letter)) word.push_back(letter);
		else cerr << "Error: letter " << letter << " not in alphabet" << endl;
	}
	void UniqueConsecutiveLetters() {
		for (int i=word.size()-1; i>=0; --i) {
			if (word[i] == word[i+1])
				word.erase(word.begin()+i);
		}
	}
	string PatternString() const {
		string s = ".*";
		for (int i=0; i<word.size(); ++i) {
			s += word[i].ToString() + ".*";
		}
		return s;
	}
	string WordString() const {
		string s;
		for (int i=0; i<word.size(); ++i) {
			s += word[i].ToString();
		}
		return s;
	}
	bool MatchPatternString(const string &ps) const {
		regex pattern(ps);
		if (regex_match(WordString(),pattern)) return true;
		return false;
	}
};


template<>
string WordType<LetterType<char>>::ToDelimitedString(char delim) const {
	string s;
	for (int i=0; i<word.size(); ++i) {
		s += word[i].ToDelimitedString();
	}
	return s;
}

template<>
string WordType<LetterType<string>>::ToDelimitedString(char delim) const {
	delim = '_';
	string s;
	for (int i=0; i<word.size(); ++i) {
		s += word[i].ToDelimitedString() + delim;
	}
	if (s.size() > 0) s.erase(s.size()-1);
	return s;
}

template<>
string WordType<LetterType<int>>::ToDelimitedString(char delim) const {
	delim = '_';
	string s;
	for (int i=0; i<word.size(); ++i) {
		s += word[i].ToDelimitedString() + delim;
	}
	if (s.size() > 0) s.erase(s.size()-1);
	return s;
}

template<class Type>
inline ostream& operator<< (ostream &os, const WordType<Type> &s) {
	os << s.ToString();
	return os;
}

template<class Type>
class P6Type {
//Listing all paths
protected:
	Alphabet<LetterType<Type> > alphabet; 
	set<WordType<LetterType<Type> > > words;
public:
	const static P6Type zero;
	const static P6Type one;
	P6Type()  {}
	P6Type(const Alphabet<LetterType<Type> > &a) : alphabet (a) {}
	P6Type(const WordType<LetterType<Type> > &w) {
		Insert(w);
	}
	P6Type(const Alphabet<LetterType<Type> > &a, const WordType<LetterType<Type> > &w) : alphabet (a) {
		Insert(w);
	}
	const P6Type& operator+ (const P6Type &s) const {
		static P6Type *r = new P6Type(alphabet);
		r->words.clear();
		for (auto i = words.begin(); i !=words.end(); ++i ) {
			r->Insert(*i);
		}
		for (auto i = s.words.begin(); i !=s.words.end(); ++i ) {
			r->Insert(*i);
		}
		return *r;
	}
	const P6Type& operator* (const P6Type &s) const {
		static P6Type *r = new P6Type(alphabet);
		r->words.clear();
		for (auto i = words.begin(); i !=words.end(); ++i ) {
			for (auto j = s.words.begin(); j !=s.words.end(); ++j ) {
				r->Insert((*i) * (*j));
			}
		}
		return *r;
	}
	bool operator==(const P6Type &s) const {
		if (words == s.words) return true; 
		return false;
	}
	bool operator<=(const P6Type &s) const {
		return   ( ((*this) + s) == s ); 
	}

	string ToString() const {
		string s = "{";
		for (auto i = words.begin(); i != words.end(); ++i ) {
			s += i->ToString();
			s += ',';
		}
		if (s[s.size()-1] == ',') s[s.size()-1] = '}';
		else s += "}";
		return s;
	}

	string ToDelimitedString(char delim = ' ') const {
		string s;
		for (auto i = words.begin(); i != words.end(); ++i ) {
			s += i->ToDelimitedString() + delim;
		}
		return s;
	}

	void FromString(string &s) {
		auto it =  find_if( s.begin() , s.end() , [](char ch){ return !isspace<char>(ch , locale::classic() ) ; } );
  		s.erase( s.begin() , it);
  		if (s[0] == '{') s = s.substr(1) ;
		else exit(0);
		istringstream is (s);
		WordType<LetterType<Type> > word;
		word.FromString(s) ;
		Insert(word);
		it =  find_if( s.begin() , s.end() , [](char ch){ return !isspace<char>(ch , locale::classic() ) ; } );
  		s.erase( s.begin() , it);
  		if (s[0] == '}') s = s.substr(1) ;
		else exit(0);
	}

	void FromDelimitedString(string &s, char delim = ' ') {
		auto it =  find_if( s.begin() , s.end() , [](char ch){ return !isspace<char>(ch , locale::classic() ) ; } );
  		s.erase( s.begin() , it);
		istringstream is (s);
		ostringstream os;
		LetterType<Type> letter;
		letter.FromString(s) ;
		WordType<LetterType<Type> > word (letter);
		Insert(word);
		it =  find_if( s.begin() , s.end() , [](char ch){ return !isspace<char>(ch , locale::classic() ) ; } );
  		s.erase( s.begin() , it);
	}

	void Insert(const WordType<LetterType<Type> > &w)  {
		words.insert(w);
	}

	const P6Type& Zero() const {
		return zero;
		
	}
	const P6Type& One() const {
		return one;
	}
	void bas() {
	}

};

template<class Type>
using PAType = P6Type<Type>;

template<class Type>
inline ostream& operator<< (ostream &os, const P6Type<Type> &s) {
	os << s.ToString();
	return os;
}


template<class Type>
class P8Type  {
//Listing elementary paths
	Alphabet<LetterType<Type> > alphabet; 
	set<WordType<LetterType<Type> > > words;
public:
	const static P8Type zero;
	const static P8Type one;
	P8Type() {}
	P8Type(const Alphabet<LetterType<Type> > &a) : alphabet(a) {}
	P8Type(const WordType<LetterType<Type> > &w) {
		Insert(w);
	}
	P8Type(const Alphabet<LetterType<Type> > &a, const WordType<LetterType<Type> > &w) : alphabet (a) {
		Insert(w);
	}
	const P8Type& operator+ (const P8Type &s) const {
		static P8Type *r = new P8Type(alphabet);
		r->words.clear();
		for (auto i = words.begin(); i !=words.end(); ++i ) {
			r->Insert(*i); 
		}
		for (auto i = s.words.begin(); i !=s.words.end(); ++i ) {
			r->Insert(*i);
		}
		return *r;
	}
	const P8Type& operator* (const P8Type &s) const {
		static P8Type *r = new P8Type(alphabet);
		r->words.clear();
		for (auto i = words.begin(); i !=words.end(); ++i ) {
			for (auto j = s.words.begin(); j !=s.words.end(); ++j ) {
				r->words.insert((*i) * (*j)); 
			}
		}
		return *r;
	}
	bool operator==(const P8Type &s) const {
		if (words == s.words) return true; 
		return false;
	}
	bool operator<=(const P8Type &s) const {
		return   ( ((*this) + s) == s ); 
	}
	string ToString() const {
		string s = "{";
		for (auto i = words.begin(); i != words.end(); ++i ) {
			s += i->ToString();
			s += ',';
		}
		if (s[s.size()-1] == ',') s[s.size()-1] = '}';
		else s += "}";
		return s;
	}

	string ToDelimitedString(char delim = ' ') const {
		string s;
		for (auto i = words.begin(); i != words.end(); ++i ) {
			s += i->ToDelimitedString() + delim;
		}
		return s;
	}

	void FromString(string &s) {
		auto it =  find_if( s.begin() , s.end() , [](char ch){ return !isspace<char>(ch , locale::classic() ) ; } );
  		s.erase( s.begin() , it);
  		if (s[0] == '{') s = s.substr(1) ;
		else exit(0);
		istringstream is (s);
		WordType<LetterType<Type> > word;
		word.FromString(s) ;
		Insert(word);
		it =  find_if( s.begin() , s.end() , [](char ch){ return !isspace<char>(ch , locale::classic() ) ; } );
  		s.erase( s.begin() , it);
  		if (s[0] == '}') s = s.substr(1) ;
		else exit(0);
	}

	void FromDelimitedString(string &s, char delim = ' ') {
		auto it =  find_if( s.begin() , s.end() , [](char ch){ return !isspace<char>(ch , locale::classic() ) ; } );
  		s.erase( s.begin() , it);
		istringstream is (s);
		LetterType<Type> letter;
		letter.FromString(s) ;
		WordType<LetterType<Type> > word (letter);
		Insert(word);
		it =  find_if( s.begin() , s.end() , [](char ch){ return !isspace<char>(ch , locale::classic() ) ; } );
  		s.erase( s.begin() , it);
	}

	const P8Type& Zero() const {
		return zero;
		
	}
	const P8Type& One() const {
		return one;
	}
	void bas() {
	}

	void Insert(const WordType<LetterType<Type> > &w)  {
		string ws = w.WordString();
		string wp = w.PatternString();
		for (auto i = words.begin(); i != words.end(); ) {
			string is = i->WordString();
			string ip = i->PatternString();
			if (is.size() <= ws.size()) {
				if ( w.MatchPatternString( ip ) ) {
					return;
				}
				else {
					++i;
				}
			}
			else { 
				if ( i->MatchPatternString( wp ) ) {
					i = words.erase(i);
				}
				else {
					++i;
				}
			}
			if (words.empty()) {
				break;
			}
		}
		words.insert(w);
	}
};


template<class Type>
using PEType = P8Type<Type>;

template<class Type>
inline ostream& operator<< (ostream &os, const P8Type<Type> &s) {
	os << s.ToString();
	return os;
}



const P1Type P1Type::zero(true) ;
const P1Type P1Type::one (false) ;

template<class Type>
const P2Type<Type> P2Type<Type>::zero(numeric_limits<Type>::infinity()) ;
template<class Type>
const P2Type<Type> P2Type<Type>::one (0.0) ;

template<class Type>
const P3Type<Type> P3Type<Type>::zero(-numeric_limits<Type>::infinity()) ;
template<class Type>
const P3Type<Type> P3Type<Type>::one (0.0) ;

const P4Type P4Type::zero(0.0) ;
const P4Type P4Type::one (1.0) ;

template<class Type>
const P5Type<Type> P5Type<Type>::zero(0.0) ;
template<class Type>
const P5Type<Type> P5Type<Type>::one (numeric_limits<Type>::infinity()) ;



template<class Type>
const P6Type<Type> P6Type<Type>::zero ;
template<class Type>
const P6Type<Type> P6Type<Type>::one ( (WordType<LetterType<Type> > ( )) ) ;

template<class Type>
const P8Type<Type> P8Type<Type>::zero ;
template<class Type>
const P8Type<Type> P8Type<Type>::one ( (WordType<LetterType<Type> > ( )) ) ;


#endif

