#ifndef __TIMER__ 
#define __TIMER__ 

#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <chrono>
#include <algorithm>

using namespace std;

class Timer {
	std::chrono::time_point<std::chrono::system_clock> startTime;
	std::chrono::time_point<std::chrono::system_clock> endTime;
	string message;
	bool coutOnDestruction;
public:
	Timer (string m = "Execution time ", bool cod = true) : coutOnDestruction(cod), message(m) {
		startTime = now();
		endTime = startTime;
	}
	~Timer () {
		if (coutOnDestruction) cout << ToString() << endl;
	}
	inline void start (string m = "Execution time ", bool cod = true) {
		message = m;
		coutOnDestruction = cod;
		startTime = now();
	}
	inline void end () {
		endTime = now();
	}
	inline std::chrono::time_point<std::chrono::system_clock> now () const {
		return chrono::high_resolution_clock::now();
	}
	inline auto ns () const {
		return chrono::duration_cast<chrono::nanoseconds>( endTime - startTime ).count();
	}
	inline auto mcs () const {
		return chrono::duration_cast<chrono::microseconds>( endTime - startTime ).count();
	}
	inline auto ms () const {
		return chrono::duration_cast<chrono::milliseconds>( endTime - startTime ).count();
	}
	inline auto s () const {
		return chrono::duration_cast<chrono::seconds>( endTime - startTime ).count();
	}
	inline auto m () const {
		return chrono::duration_cast<chrono::minutes>( endTime - startTime ).count();
	}
	inline auto h () const {
		return chrono::duration_cast<chrono::hours>( endTime - startTime ).count();
	}
	string ToString () const {
		ostringstream os;
		os << message << ms() << " ms";
		return os.str();
	}
};

ostream& operator<< (ostream &os, const Timer &t) {
	os << t.ToString();
	return os;
}

void PrintError(const string &s) {
	cerr << s << endl;
}

void PrintErrorAndExit(const string &s) {
	PrintError(s);
	exit(1);
}

void EraseLeadingBlanks(string &s) {
	s.erase( s.begin(), find_if( s.begin() , s.end() , [](char ch){ return !isspace<char>(ch , locale::classic() ) ; } ));
}

void EraseComments(string &s, const string &comment = "//") {
	auto pos = s.find(comment);
	if (pos != string::npos) s.erase( pos );
}

void EraseReadCharacters(istringstream &is) {
	if (is.fail()) PrintErrorAndExit("EraseReadCharacters Error: cannot read from:" + is.str() + ".");
	if (is.tellg() > 0) is.str(is.str().substr(is.tellg()));
	else is.str("");
}

void PrintToFile(const string &fileName, const string &message) {
	ofstream f (fileName);
	if (!f.is_open()) PrintErrorAndExit("PrintToFile Error: cannot open file.");
	f << message << endl;
	f.close();
}

void AppendToFile(const string &fileName, const string &message) {
	ofstream os (fileName, std::ofstream::app);
	if (!os.is_open()) PrintErrorAndExit("PrintToFile Error: cannot open file.");
	os << message << endl;
	os.close();
}

class ProblemParameters {
public:
	string folderfile;
	string folder;
	string filename;
	string t; //problem type: SP, MRP ...
	int n; //n: number of graph nodes
	double d; // graph density
	int s; //number of scenarios
	int k; //problem instance number (1 to 30)
	string solutionFileHeader;
	ProblemParameters(const string &ff) {
		ParseFileName(ff);
	}
	ProblemParameters(char **argv, int argc) {
		if (argc > 1) ParseFileName(argv[1]);
		else PrintErrorAndExit("ProblemParameters Error: invalid argc.");
	}
	void ParseFileName(const string &ff) {
		folderfile = ff;
		folder = folderfile.substr(0,folderfile.find_last_of('\\')+1);
		filename = folderfile.substr(folderfile.find_last_of('\\')+1);
		t = filename.substr(0,filename.find_first_of('-')-1); 
		n = atoi(filename.substr(filename.find("-n")+2).c_str());
		d = atof(filename.substr(filename.find_last_of('c')+1).c_str());
		s = atoi(filename.substr(filename.find_last_of('s')+1).c_str());
		k = atoi(filename.substr(filename.find_last_of('-')+1).c_str());
		ostringstream solutionFileHeaderStream;
			solutionFileHeaderStream << "//" << filename << " solution file:\n";
			solutionFileHeaderStream << "//\t" << "t" << "\t" << "n" << "\t" << "d" << "\t" << "s" << "\t" << "k" << "\n";
			solutionFileHeaderStream << "//\t" << t << "\t" << n << "\t" << d << "\t" << s << "\t" << k; // << "\n";
		solutionFileHeader = solutionFileHeaderStream.str();
	}
};

#endif

